from abc import ABCMeta
from abc import abstractmethod
from collections import namedtuple
from contextlib import contextmanager
from os import remove
from tarfile import TarFile
from tempfile import NamedTemporaryFile
from typing import IO
from typing import Iterable
from typing import Tuple
from typing import TypeVar
from uuid import uuid4

from cached_property import cached_property
from libcloud.storage.base import Container
from libcloud.storage.providers import Provider
from libcloud.storage.providers import get_driver
from libcloud.storage.types import ObjectDoesNotExistError
from xtarfile import open as tarfile_open

from opwen_email_client.domain.email.client import EmailServerClient
from opwen_email_client.domain.email.user_store import User
from opwen_email_client.util.serialization import Serializer

T = TypeVar('T')

Download = namedtuple('Download', ('name', 'optional', 'type_'))


class Sync(metaclass=ABCMeta):
    @abstractmethod
    def upload(self, items: Iterable[T], users: Iterable[User]) -> Iterable[str]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def download(self) -> Iterable[T]:
        raise NotImplementedError  # pragma: no cover


class AzureSync(Sync):
    _emails_file = 'emails.jsonl'
    _attachments_file = 'zattachments.jsonl'
    _users_file = 'zzusers.jsonl'

    _download_files = (
        Download(name=_emails_file, optional=False, type_='email'),
        Download(name=_attachments_file, optional=True, type_='attachment'),
    )

    def __init__(self, container: str, serializer: Serializer, account_name: str, account_key: str,
                 email_server_client: EmailServerClient, provider: str, compression: str):

        self._container = container
        self._serializer = serializer
        self._account = account_name
        self._key = account_key
        self._email_server_client = email_server_client
        self._provider = getattr(Provider, provider)
        self._compression = compression

    @cached_property
    def _azure_client(self) -> Container:
        driver = get_driver(self._provider)
        client = driver(self._account, self._key)
        return client.get_container(self._container)

    @classmethod
    @contextmanager
    def _workspace(cls, suffix: str) -> IO:
        temp = NamedTemporaryFile(suffix=suffix, delete=False)
        try:
            temp.close()
            fobj = open(temp.name, mode=temp.mode)
            try:
                yield fobj
            finally:
                fobj.close()
        finally:
            remove(temp.name)

    def _open(self, path: str, mode: str) -> TarFile:
        extension_index = path.rfind('.')
        if extension_index > -1:
            compression = path[extension_index + 1:]
        else:
            compression = self._compression

        mode = '{}|{}'.format(mode, compression)

        return tarfile_open(path, mode=mode)

    def _download_to_stream(self, blobname: str, stream: IO) -> bool:

        try:
            resource = self._azure_client.get_object(blobname)
        except ObjectDoesNotExistError:
            return False
        else:
            for chunk in resource.as_stream():
                stream.write(chunk)
            return True

    def _upload_from_stream(self, blobname: str, stream: IO):
        self._azure_client.upload_object_via_stream(stream, blobname)

    @classmethod
    def _get_file_from_download(cls, archive, downloads: Iterable[Download]) \
            -> Iterable[Tuple[Download, IO[bytes]]]:

        downloads = {download.name: download for download in downloads}

        while True:
            member = archive.next()
            if member is None:
                break

            filename = member.name
            try:
                download = downloads[filename]
            except KeyError:
                pass
            else:
                yield download, archive.extractfile(member)
                del downloads[filename]

        missing_downloads = [filename for filename, download in downloads.items() if not download.optional]

        if missing_downloads:
            raise FileNotFoundError(','.join(missing_downloads))

    def download(self):
        resource_id = self._email_server_client.download()
        if not resource_id:
            return

        with self._workspace(resource_id) as workspace:
            if not self._download_to_stream(resource_id, workspace):
                return

            workspace.seek(0)
            with self._open(workspace.name, 'r') as archive:
                for download, fobj in self._get_file_from_download(archive, self._download_files):
                    for line in fobj:
                        obj = self._serializer.deserialize(line, download.type_)
                        obj['_type'] = download.type_
                        yield obj

    def _upload_emails(self, items, archive):
        uploaded_ids = []

        with self._workspace(self._emails_file) as uploaded:
            for item in items:
                item = {key: value for (key, value) in item.items() if value is not None}
                item.pop('read', False)
                for attachment in item.get('attachments', []):
                    attachment.pop('_uid', '')
                serialized = self._serializer.serialize(item)
                uploaded.write(serialized)
                uploaded.write(b'\n')
                uploaded_ids.append(item.get('_uid'))

            uploaded.seek(0)
            archive.add(uploaded.name, self._emails_file)

        return uploaded_ids

    def _upload_users(self, users, archive):
        if not users:
            return

        with self._workspace(self._users_file) as uploaded:
            for user in users:
                item = {'email': user.email, 'password': user.password}
                serialized = self._serializer.serialize(item)
                uploaded.write(serialized)
                uploaded.write(b'\n')

            uploaded.seek(0)
            archive.add(uploaded.name, self._users_file)

    def upload(self, items, users):
        upload_location = '{}.tar.{}'.format(uuid4(), self._compression)

        with self._workspace(upload_location) as workspace:
            with self._open(workspace.name, 'w') as archive:
                uploaded_ids = self._upload_emails(items, archive)
                self._upload_users(users, archive)

            if uploaded_ids:
                workspace.seek(0)
                self._upload_from_stream(upload_location, workspace)
                self._email_server_client.upload(upload_location, self._container)

        return uploaded_ids



from importlib import import_module

from cached_property import cached_property
from flask import Flask
from flask_babelex import Babel

from opwen_email_client.domain.email.client import HttpEmailServerClient
from opwen_email_client.domain.email.client import LocalEmailServerClient
from opwen_email_client.domain.email.sql_store import SqliteEmailStore
from opwen_email_client.domain.email.sync import AzureSync
from opwen_email_client.util.serialization import JsonSerializer
from opwen_email_client.webapp.cache import cache
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.forms.login import LoginForm
from opwen_email_client.webapp.forms.login import RegisterForm
from opwen_email_client.webapp.login import FlaskLoginUserStore
from opwen_email_client.webapp.mkwvconf import blueprint as mkwvconf
from opwen_email_client.webapp.security import security


class Ioc:
    @cached_property
    def email_store(self):
        return SqliteEmailStore(
            page_size=AppConfig.EMAILS_PER_PAGE,
            restricted={AppConfig.NEWS_INBOX: AppConfig.NEWS_SENDERS},
            database_path=AppConfig.LOCAL_EMAIL_STORE,
        )

    @cached_property
    def email_sync(self):
        if AppConfig.TESTING:
            email_server_client = LocalEmailServerClient()
        else:
            email_server_client = HttpEmailServerClient(
                compression=AppConfig.COMPRESSION,
                hostname=AppConfig.EMAIL_SERVER_HOSTNAME,
                client_id=AppConfig.CLIENT_ID,
            )

        serializer = JsonSerializer()

        return AzureSync(
            compression=AppConfig.COMPRESSION,
            account_name=AppConfig.STORAGE_ACCOUNT_NAME,
            account_key=AppConfig.STORAGE_ACCOUNT_KEY,
            email_server_client=email_server_client,
            container=AppConfig.STORAGE_CONTAINER,
            provider=AppConfig.STORAGE_PROVIDER,
            serializer=serializer,
        )

    @cached_property
    def user_store(self):
        return FlaskLoginUserStore()

    @cached_property
    def login_form(self):
        return LoginForm


def _new_ioc(fqn: str) -> Ioc:
    fqn_parts = fqn.split('.')
    class_name = fqn_parts.pop()
    module_name = '.'.join(fqn_parts)

    module = import_module(module_name)
    cls = getattr(module, class_name)

    return cls()


def create_app(config=AppConfig) -> Flask:
    app = Flask(__name__, static_url_path=config.APP_ROOT + '/static')
    app.config.from_object(config)

    app.babel = Babel(app)

    app.ioc = _new_ioc(config.IOC)

    cache.init_app(app)
    app.ioc.user_store.init_app(app)
    security.init_app(app, app.ioc.user_store.r, register_form=RegisterForm, login_form=app.ioc.login_form)

    app.register_blueprint(mkwvconf, url_prefix=config.APP_ROOT + '/api/mkwvconf')

    return app



from flask_login import login_required as _login_required
from flask_security import Security

from opwen_email_client.webapp.config import AppConfig

security = Security()


def login_required(func):
    if AppConfig.TESTING:
        return func

    return _login_required(func)



from datetime import datetime
from datetime import timedelta
from io import BytesIO
from os import path
from typing import Iterable
from typing import Optional

from babel import Locale
from flask import Response
from flask import abort
from flask import flash
from flask import jsonify
from flask import redirect
from flask import render_template
from flask import request
from flask import send_file
from flask import send_from_directory
from flask import url_for
from flask_login import current_user
from flask_security.utils import hash_password
from passlib.pwd import genword

from opwen_email_client.webapp import app
from opwen_email_client.webapp import tasks
from opwen_email_client.webapp.actions import SendWelcomeEmail
from opwen_email_client.webapp.config import AppConfig
from opwen_email_client.webapp.config import i8n
from opwen_email_client.webapp.forms.email import NewEmailForm
from opwen_email_client.webapp.forms.settings import SettingsForm
from opwen_email_client.webapp.security import login_required
from opwen_email_client.webapp.session import Session
from opwen_email_client.webapp.session import track_history


@app.route(AppConfig.APP_ROOT + '/favicon.ico')
def favicon() -> Response:
    return send_from_directory(
        directory=path.join(app.root_path, 'static'),
        filename='favicon.ico',
        mimetype='image/vnd.microsoft.icon',
    )


@app.route(AppConfig.APP_ROOT + '/')
@app.route(AppConfig.APP_ROOT + '/home')
@track_history
def home() -> Response:
    return _view('home.html')


@app.route(AppConfig.APP_ROOT + '/about')
@track_history
def about() -> Response:
    return _view('about.html')


@app.route(AppConfig.APP_ROOT + '/news', defaults={'page': 1})
@app.route(AppConfig.APP_ROOT + '/news/<int:page>')
@track_history
def news(page: int) -> Response:
    email_store = app.ioc.email_store

    return _emails_view(email_store.inbox(AppConfig.NEWS_INBOX, page), page, 'news.html')


@app.route(AppConfig.APP_ROOT + '/email')
@app.route(AppConfig.APP_ROOT + '/email/inbox', defaults={'page': 1})
@app.route(AppConfig.APP_ROOT + '/email/inbox/<int:page>')
@login_required
@track_history
def email_inbox(page: int) -> Response:
    email_store = app.ioc.email_store
    user = current_user

    return _emails_view(email_store.inbox(user.email, page), page, type='inbox')


@app.route(AppConfig.APP_ROOT + '/email/outbox', defaults={'page': 1})
@app.route(AppConfig.APP_ROOT + '/email/outbox/<int:page>')
@login_required
@track_history
def email_outbox(page: int) -> Response:
    email_store = app.ioc.email_store
    user = current_user

    return _emails_view(email_store.outbox(user.email, page), page, type='outbox')


@app.route(AppConfig.APP_ROOT + '/email/sent', defaults={'page': 1})
@app.route(AppConfig.APP_ROOT + '/email/sent/<int:page>')
@login_required
@track_history
def email_sent(page: int) -> Response:
    email_store = app.ioc.email_store
    user = current_user

    return _emails_view(email_store.sent(user.email, page), page, type='sent')


@app.route(AppConfig.APP_ROOT + '/email/search', defaults={'page': 1})
@app.route(AppConfig.APP_ROOT + '/email/search/<int:page>')
@login_required
@track_history
def email_search(page: int) -> Response:
    if not AppConfig.EMAIL_SEARCHABLE:
        return abort(404)

    email_store = app.ioc.email_store
    user = current_user
    query = request.args.get('query')

    return _emails_view(email_store.search(user.email, page, query), page, 'email_search.html', type='search')


@app.route(AppConfig.APP_ROOT + '/email/unread')
@login_required
def email_unread() -> Response:
    email_store = app.ioc.email_store
    user = current_user

    return jsonify(unread=email_store.has_unread(user.email))


@app.route(AppConfig.APP_ROOT + '/email/read/<email_uid>')
@login_required
def email_read(email_uid: str) -> Response:
    email_store = app.ioc.email_store
    user = current_user

    email_store.mark_read(user.email, [email_uid])

    return Response('OK', status=200, mimetype='text/plain')


@app.route(AppConfig.APP_ROOT + '/email/delete/<email_uid>')
@login_required
def email_delete(email_uid: str) -> Response:
    email_store = app.ioc.email_store
    user = current_user

    email_store.delete(user.email, [email_uid])

    return redirect(Session.get_last_visited_url() or url_for('home'))


@app.route(AppConfig.APP_ROOT + '/email/new', methods=['GET', 'POST'])
@login_required
@track_history
def email_new() -> Response:
    email_store = app.ioc.email_store

    form = NewEmailForm.from_request(email_store)
    if form is None:
        return abort(404)

    if form.validate_on_submit():
        email_store.create([form.as_dict(email_store)])
        flash(i8n.EMAIL_SENT, category='success')
        return redirect(url_for('email_inbox'))

    return _view('email_new.html', max_upload_size_mb=AppConfig.MAX_UPLOAD_SIZE_MB, form=form, type='write')


@app.route(AppConfig.APP_ROOT + '/attachment/<email_id>/<attachment_id>', methods=['GET'])
@login_required
def download_attachment(email_id: str, attachment_id: str) -> Response:
    email_store = app.ioc.email_store

    attachment = email_store.get_attachment(email_id, attachment_id)
    if attachment is None:
        return abort(404)

    return send_file(
        BytesIO(attachment['content']),
        attachment_filename=attachment['filename'],
        as_attachment=True,
    )


@app.route(AppConfig.APP_ROOT + '/user/register/complete')
@login_required
def register_complete() -> Response:
    user_store = app.ioc.user_store

    send_welcome_email = SendWelcomeEmail(
        time=datetime.utcnow(),
        to=current_user.email,
        email_store=app.ioc.email_store,
    )

    send_welcome_email()

    current_user.language = Session.get_current_language()
    user_store.w.put(current_user)
    user_store.w.commit()

    flash(i8n.ACCOUNT_CREATED, category='success')
    return redirect(url_for('email_inbox'))


@app.route(AppConfig.APP_ROOT + '/user/login/complete')
@login_required
def login_complete() -> Response:
    current_language = current_user.language
    if current_language:
        Session.store_current_language(current_language)

    flash(i8n.LOGGED_IN, category='success')
    return redirect(url_for('home'))


@app.route(AppConfig.APP_ROOT + '/user/logout/complete')
def logout_complete() -> Response:
    flash(i8n.LOGGED_OUT, category='success')
    return redirect(url_for('home'))


@app.route(AppConfig.APP_ROOT + '/admin/sync')
def sync() -> Response:
    if not current_user.is_admin:
        abort(403)

    tasks.sync.delay()

    flash(i8n.SYNC_RUNNING, category='success')
    return redirect(url_for('settings'))


@app.route(AppConfig.APP_ROOT + '/admin/update', defaults={'version': None})
@app.route(AppConfig.APP_ROOT + '/admin/update/<version>')
def update(version: Optional[str]) -> Response:
    if not current_user.is_admin:
        abort(403)

    tasks.update.delay(version=version)

    flash(i8n.UPDATE_RUNNING, category='success')
    return redirect(url_for('settings'))


@app.route(AppConfig.APP_ROOT + '/user/language/<locale>')
def language(locale: str) -> Response:
    user_store = app.ioc.user_store

    if current_user.is_authenticated:
        current_user.language = locale
        user_store.w.put(current_user)
        user_store.w.commit()

    Session.store_current_language(locale)
    return redirect(Session.get_last_visited_url() or url_for('home'))


@app.route(AppConfig.APP_ROOT + '/users')
@login_required
@track_history
def users() -> Response:
    user_store = app.ioc.user_store

    return _view('users.html', users=user_store.fetch_all(current_user))


@app.route(AppConfig.APP_ROOT + '/admin/settings', methods=['GET', 'POST'])
@track_history
def settings() -> Response:
    if not current_user.is_admin:
        abort(403)

    email_store = app.ioc.email_store

    form = SettingsForm()
    if form.validate_on_submit():
        form.update()
        flash(i8n.SETTINGS_UPDATED, category='success')
        return redirect(url_for('settings'))

    return _view('settings.html', form=SettingsForm.from_config(), num_pending=email_store.num_pending())


@app.route(AppConfig.APP_ROOT + '/admin/suspend/<userid>')
def suspend(userid: str) -> Response:
    if not current_user.is_admin:
        abort(403)

    user_store = app.ioc.user_store

    user = user_store.r.find_user(id=userid)

    if user is None:
        flash(i8n.USER_DOES_NOT_EXIST, category='error')
        return redirect(url_for('users'))

    if user.is_admin:
        flash(i8n.ADMIN_CANNOT_BE_SUSPENDED, category='error')
        return redirect(url_for('users'))

    user.active = False
    user_store.w.put(user)
    user_store.w.commit()

    flash(i8n.USER_SUSPENDED, category='success')
    return redirect(url_for('users'))


@app.route(AppConfig.APP_ROOT + '/admin/unsuspend/<userid>')
def unsuspend(userid: str) -> Response:
    if not current_user.is_admin:
        abort(403)

    user_store = app.ioc.user_store

    user = user_store.r.find_user(id=userid)

    if user is None:
        flash(i8n.USER_DOES_NOT_EXIST, category='error')
        return redirect(url_for('users'))

    user.active = True
    user_store.w.put(user)
    user_store.w.commit()

    flash(i8n.USER_UNSUSPENDED, category='success')
    return redirect(url_for('users'))


@app.route(AppConfig.APP_ROOT + '/admin/promote/<userid>')
def promote(userid: str) -> Response:
    if not current_user.is_admin:
        abort(403)

    user_store = app.ioc.user_store

    user = user_store.r.find_user(id=userid)

    if user is None:
        flash(i8n.USER_DOES_NOT_EXIST, category='error')
        return redirect(url_for('users'))

    if user.is_admin:
        flash(i8n.ALREADY_PROMOTED, category='error')
        return redirect(url_for('users'))

    user.is_admin = True
    user_store.w.put(user)
    user_store.w.commit()

    flash(i8n.USER_PROMOTED, category='success')
    return redirect(url_for('users'))


@app.route(AppConfig.APP_ROOT + '/admin/password/reset/<userid>')
def reset_password(userid: str) -> Response:
    if not current_user.is_admin:
        abort(403)

    user_store = app.ioc.user_store

    user = user_store.r.find_user(id=userid)

    if user is None:
        flash(i8n.USER_DOES_NOT_EXIST, category='error')
        return redirect(url_for('users'))

    if user.is_admin:
        flash(i8n.ADMIN_PASSWORD_CANNOT_BE_RESET, category='error')
        return redirect(url_for('users'))

    new_password = genword()
    user.password = hash_password(new_password)
    user_store.w.put(user)
    user_store.w.commit()

    flash(i8n.PASSWORD_CHANGED_BY_ADMIN + new_password, category='success')
    return redirect(url_for('users'))


# noinspection PyUnusedLocal
@app.errorhandler(404)
def _on_404(status_code: int) -> Response:
    app.logger.warning('404: %s', request.path)
    flash(i8n.PAGE_DOES_NOT_EXIST, category='error')
    return redirect(url_for('home'))


@app.errorhandler(Exception)
def _on_exception(exception: Exception) -> Response:
    app.logger.error('%s: %s', exception.__class__.__name__, exception)
    flash(i8n.UNEXPECTED_ERROR, category='error')
    return redirect(url_for('home'))


@app.errorhandler(500)
def _on_500(status_code: int) -> Response:
    app.logger.error(status_code)
    flash(i8n.UNEXPECTED_ERROR, category='error')
    return redirect(url_for('home'))


@app.context_processor
def _inject_config() -> dict:
    return {
        'locales': AppConfig.LOCALES,
        'current_locale': Locale.parse(_localeselector()),
        'local_only': AppConfig.SIM_TYPE == 'LocalOnly',
        'app_root': AppConfig.APP_ROOT,
        'can_change_password': AppConfig.SECURITY_CHANGEABLE,
        'can_register_user': AppConfig.SECURITY_REGISTERABLE,
        'can_search_email': AppConfig.EMAIL_SEARCHABLE,
    }


@app.babel.localeselector
def _localeselector() -> str:
    current_language = Session.get_current_language()
    if not current_language and current_user.is_authenticated:
        current_language = current_user.language
    if not current_language:
        current_language = AppConfig.DEFAULT_LOCALE.language
    return current_language


def _emails_view(emails: Iterable[dict], page: int, template: str = 'email.html', **kwargs) -> Response:
    offset_minutes = getattr(current_user, 'timezone_offset_minutes', None) or 0
    timezone_offset = timedelta(minutes=offset_minutes)

    emails = list(emails)

    for email in emails:
        sent_at = email.get('sent_at')
        if sent_at:
            sent_at_utc = datetime.strptime(sent_at, '%Y-%m-%d %H:%M')
            sent_at_local = sent_at_utc - timezone_offset
            email['sent_at'] = sent_at_local.strftime('%Y-%m-%d %H:%M')

    return _view(template,
                 emails=emails,
                 page=page,
                 has_prevpage=page > 1,
                 has_nextpage=len(emails) == AppConfig.EMAILS_PER_PAGE,
                 **kwargs)


def _view(template: str, **kwargs) -> Response:
    return render_template(template, **kwargs)



from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Union

from flask import Flask
from flask_login import UserMixin
from flask_security.datastore import Datastore as UserWriteStore
from flask_security.datastore import UserDatastore as UserReadStore


class User(UserMixin):
    @property
    @abstractmethod
    def id(self) -> Union[str, int]:
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def email(self) -> str:
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def password(self) -> str:
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def roles(self) -> List[str]:
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def active(self) -> bool:
        raise NotImplementedError  # pragma: no cover


class UserStore(metaclass=ABCMeta):
    def __init__(self, read: UserReadStore, write: UserWriteStore) -> None:
        self.r = read
        self.w = write

    def init_app(self, app: Flask) -> None:
        pass

    @abstractmethod
    def fetch_all(self, user: User) -> List[User]:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def fetch_pending(self) -> List[User]:
        raise NotImplementedError  # pragma: no cover



import re
from pathlib import Path

from setuptools import find_packages
from setuptools import setup

package = 'opwen_email_client'
here = Path(__file__).parent

with (here / 'requirements.txt').open(encoding='utf-8') as fobj:
    install_requires = [line.strip() for line in fobj]

with (here / 'README.rst').open(encoding='utf-8') as fobj:
    long_description = fobj.read()

with (here / package / '__init__.py').open(encoding='utf-8') as fobj:
    version = re.search(r"^__version__ = '([^']*)'",
                        fobj.read(), re.MULTILINE).group(1)

scripts = ['manage.py']

setup(
    name=package,
    version=version,
    author='Clemens Wolff',
    author_email='clemens.wolff+pypi@gmail.com',
    packages=find_packages(exclude=['tests*']),
    url='https://github.com/ascoderu/opwen-webapp',
    license='Apache Software License',
    description='Email client for the Lokole project: https://ascoderu.ca',
    long_description=long_description,
    include_package_data=True,
    scripts=scripts,
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Topic :: Communications :: Email',
    ])
